@@ -4,7 +4,7 @@ with OPTIMADE providers that can be used in server or client code.
 """
 
 import json
-from typing import Iterable, List
+from typing import Container, Iterable, List, Optional
 
 from pydantic import ValidationError
 
@@ -101,7 +101,7 @@ def get_providers(add_mongo_id: bool = False) -> list:
 
 
 def get_child_database_links(
-    provider: LinksResource, obey_aggregate=True
+    provider: LinksResource, obey_aggregate: bool = True
 ) -> List[LinksResource]:
     """For a provider, return a list of available child databases.
 
@@ -155,13 +155,37 @@ def get_child_database_links(
         ) from exc
 
 
-def get_all_databases() -> Iterable[str]:
-    """Iterate through all databases reported by registered OPTIMADE providers."""
+def get_all_databases(
+    include_providers: Optional[Container[str]] = None,
+    exclude_providers: Optional[Container[str]] = None,
+    exclude_databases: Optional[Container[str]] = None,
+) -> Iterable[str]:
+    """Iterate through all databases reported by registered OPTIMADE providers.
+
+    Parameters:
+        include_providers: A set/container of provider IDs to include child databases for.
+        exclude_providers: A set/container of provider IDs to exclude child databases for.
+        exclude_databases: A set/container of specific database URLs to exclude.
+
+    Returns:
+        A generator of child database links that obey the given parameters.
+
+    """
     for provider in get_providers():
+        if exclude_providers and provider["id"] in exclude_providers:
+            continue
+        if include_providers and provider["id"] not in include_providers:
+            continue
+
         try:
             links = get_child_database_links(provider)
             for link in links:
                 if link.attributes.base_url:
+                    if (
+                        exclude_databases
+                        and link.attributes.base_url in exclude_databases
+                    ):
+                        continue
                     yield str(link.attributes.base_url)
         except RuntimeError:
             pass
